-------------------------------------------------------------------------------
--
-- TV to VGA converter
--
-- Modified from de2_raster_vga.vhd
-- by Stephen A. Edwards, Columbia University, sedwards@cs.columbia.edu
-- using YCbCr2RGB.v and itu_r656_decoder.v from Terasic Technology, Inc.
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity TV_to_VGA is
  
  port (
    reset,
    OSC_27,
	OSC_50: in std_logic;

    read       : in  std_logic;
    write      : in  std_logic;
    chipselect : in  std_logic;
    address    : in  std_logic_vector(4 downto 0);
    readdata   : out std_logic_vector(15 downto 0);
    writedata  : in  std_logic_vector(15 downto 0);

    -- SRAM
    SRAM_DQ : inout std_logic_vector(15 downto 0);	-- Data bus 16 Bits
    SRAM_ADDR : out std_logic_vector(17 downto 0);	-- Address bus 18 Bits
    SRAM_UB_N,                                    	-- High-byte Data Mask 
    SRAM_LB_N,                                     	-- Low-byte Data Mask 
    SRAM_WE_N,                                     	-- Write Enable
    SRAM_CE_N,                                     	-- Chip Enable
    SRAM_OE_N : out std_logic;                     	-- Output Enable

		-- TV
    TD_D	: in std_logic_vector(7 downto 0);				-- Data bus 8 bits
    TD_HS,                                      		-- H_SYNC
    TD_VS : in std_logic;                       		-- V_SYNC

		-- VGA
    VGA_CLOCK,																			-- Clock
    VGA_HS,                          								-- H_SYNC
    VGA_VS,                          								-- V_SYNC
    VGA_BLANK,                       								-- BLANK
    VGA_SYNC : out std_logic;        								-- SYNC
    VGA_R,                       	    							-- Red[9:0]
    VGA_G,                          		 						-- Green[9:0]
    VGA_B : out std_logic_vector(9 downto 0); 			-- Blue[9:0]

		-- Control signals
	  freeze_b, full_screen_b,
		send_image, del_image,
		move_right, move_left : in std_logic
	
    );

end TV_to_VGA;

architecture rtl of TV_to_VGA is

	-- Constants
	
	-- VGA constants
	constant vga_x_size 			: integer := 740;
	constant vga_y_size				: integer := 504;
	
	-- Screen constants
	constant screen_x_offset	: integer := 60;--97;
	constant screen_y_offset	: integer := 28;--51;
	constant screen_x_size		: integer := vga_x_size - screen_x_offset;
	constant screen_y_size		: integer := vga_y_size - screen_y_offset - 8;
	
	-- Image constants
	constant image_x_size		: integer := screen_x_size / 2;
	constant image_y_size		: integer := screen_y_size / 2;
	constant image_x_offset	: integer := (screen_x_size - image_x_size) / 2;
	constant image_y_offset	: integer := 60;
	constant image_border		: integer := 2;
	
	-- Image list constants
	constant image_list_offset	: integer := 0;
	constant image_list_border	: integer := 0;
	constant image_list_spacing	: integer := 0;
	
	-- Signals
	
	
	-- VGA counters
	signal vga_x_count, vga_y_count: unsigned(9 downto 0);
	signal EndOfLine, EndOfField : std_logic;
	signal x_reset, y_reset : std_logic;
	
	-- Screen counters
	signal screen_x_count, screen_y_count: unsigned(9 downto 0);
	signal is_within_screen	: std_logic;
	
	-- Image signal
	signal image_x_count, image_y_count: unsigned(9 downto 0);
	signal is_within_image	: std_logic;
	signal is_image_frame	: std_logic;
	signal is_within_image_list	: std_logic;
	
	-- Image list signals
	signal image_list_x_count, image_list_y_count: unsigned(9 downto 0);
	
	-- SRAM signals
	signal byteenable : std_logic_vector(1 downto 0);
	signal write_data, read_data : std_logic_vector(15 downto 0);
	signal ram_addr, image_addr, screen_addr : unsigned(17 downto 0);
	signal get_image_addr, save_image_addr : unsigned(17 downto 0);
	signal addr_tmp : unsigned(19 downto 0);
	signal sram_data, req_data : std_logic_vector(15 downto 0);
	
	-- VGA signals
	signal red_r, green_r, blue_r : std_logic_vector(9 downto 0);
	signal VGA_HS_r, VGA_VS_r : std_logic;
	
	-- Sending image signals
	signal save_image_x, save_image_y: unsigned(9 downto 0);
	signal got_x, got_y, sent_data : std_logic := '0';
	
	-- Receiving image signals
	signal getting_image : std_logic := '0';
	signal pixel : unsigned(15 downto 0);
	signal get_image_x, get_image_y : unsigned(9 downto 0);
	signal got_x2, got_y2, got_data : std_logic := '0';
	signal current_image : unsigned(9 downto 0);
	signal num_images : unsigned(9 downto 0) := "00" & x"00";

	-- Receivig thumbnails signals
	
	
	-- 
	signal save_image, addr0, write_to_ram: std_logic := '0';
	signal red_pixel, green_pixel, blue_pixel	: std_logic_vector(9 downto 0);
	signal freeze_k, full_screen_k, freeze, full_screen : std_logic;
	
	
	-- Components

	-- itu_r656_decoder and YCbCr2RGB Signals
	signal Y, Cb, Cr : std_logic_vector(7 downto 0);		-- 4:4:4 signals
	signal mTD_HSx2 : std_logic;
	signal Red, Green, Blue : std_logic_vector(9 downto 0);

	component itu_r656_decoder port(
		CLOCK	: in std_logic;
		TD_D	: in std_logic_vector(7 downto 0);
 		TD_HS,
		TD_VS	: in std_logic;
		Y,
		Cb,
		Cr		: out std_logic_vector(7 downto 0);
    HSx2,
		blank	: out std_logic);
	end component;
		
	component YCbCr2RGB port(	
		-- inptus:
		iY,
		iCb,
		iCr		: in std_logic_vector(7 downto 0);
		iRESET,
		iCLK	: in std_logic;
		-- outputs:
		Red,
		Green,
		Blue	: out std_logic_vector(9 downto 0)
		);
	end component;
	
	component VGA_Synchronization port(	
		-- inputs:
		OSC_27,
		reset,
		TD_HS,
		TD_VS,
		mTD_HSx2	: in std_logic;
		-- outputs:
		VGA_SYNC,
		VGA_CLOCK,
		VGA_HS,
		VGA_VS		: out std_logic;
		h_reset,
		v_reset		: out std_logic);
	end component;
	
	component sram port (
		-- inputs:
		address : IN STD_LOGIC_VECTOR (17 DOWNTO 0);
		byteenable : IN STD_LOGIC_VECTOR (1 DOWNTO 0);
		chipselect : IN STD_LOGIC;
		read : IN STD_LOGIC;
		write : IN STD_LOGIC;
		writedata : IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		-- outputs:
		SRAM_ADDR : OUT STD_LOGIC_VECTOR (17 DOWNTO 0);
		SRAM_CE_N : OUT STD_LOGIC;
		SRAM_DQ		: INOUT STD_LOGIC_VECTOR (15 DOWNTO 0);
		SRAM_LB_N : OUT STD_LOGIC;
		SRAM_OE_N : OUT STD_LOGIC;
		SRAM_UB_N : OUT STD_LOGIC;
		SRAM_WE_N : OUT STD_LOGIC;
		readdata	: OUT STD_LOGIC_VECTOR (15 DOWNTO 0));
	end component;
	
	
begin
	freeze <= freeze_b or freeze_k;
	full_screen <= not full_screen_b or full_screen_k;


	-- Set vga_x_count and vga_y_count
	
  EndOfLine <= '1' when vga_x_count = vga_x_size - 1 else '0';
  EndOfField <= '1' when vga_y_count = vga_y_size - 1 else '0';

  process (OSC_27)
  begin
    if rising_edge(OSC_27) then      
      if reset = '1' or VGA_HS_r = '0' then
        vga_x_count <= (others => '0');
     elsif EndOfLine = '1' then
       vga_x_count <= (others => '0');
      else
        vga_x_count <= vga_x_count + 1;
      end if;      
    end if;
  end process;

  process (OSC_27)
  begin
    if rising_edge(OSC_27) then      
      if reset = '1' or VGA_VS_r = '0' then
        vga_y_count <= (others => '0');
      elsif EndOfLine = '1' then
        if EndOfField = '1' then
          vga_y_count <= (others => '0');
        else
          vga_y_count <= vga_y_count + 1;
        end if;
      end if;
    end if;
  end process;

	-- Sub-components
	
	-- ITU Decoder
	itu: itu_r656_decoder port map(
		CLOCK	=> OSC_27,
		TD_D	=> TD_D,
 		TD_HS	=> TD_HS,
		TD_VS	=> TD_VS,
		Y		=> Y,	
		Cb		=> Cb,	
		Cr		=> Cr,	
    	HSx2	=> mTD_HSx2,
		blank	=> VGA_BLANK
	);
	
	-- Color Space Convertor (YCbCr to RGB)
	color_space_convertor: YCbCr2RGB port map(	
		Red		=> red_r,
		Green	=> green_r,
		Blue	=> blue_r,
		iY		=> Y,
		iCb		=> Cb,
		iCr		=> Cr,			
		iRESET	=> RESET,
		iCLK	=> OSC_27
	);

	-- Synchronization
	sync: VGA_Synchronization port map(
		OSC_27		=> OSC_27,
		reset		=> reset,
		TD_HS		=> TD_HS,
		TD_VS		=> TD_VS,
		mTD_HSx2	=> mTD_HSx2,
		VGA_SYNC	=> VGA_SYNC,
		VGA_CLOCK	=> VGA_CLOCK,
		VGA_HS		=> VGA_HS_r,
		VGA_VS		=> VGA_VS_r,
		h_reset		=> x_reset,
		v_reset		=> y_reset
	);

	VGA_HS <= VGA_HS_r;
	VGA_VS <= VGA_VS_r;
	
	-- Send pixel value when requested
	-- address:
	--		0	- receive x position
	--		1	- receive y position
	--		2 - get save_image signal
	--		3 - send data
	process (OSC_50)
	begin
		if rising_edge(OSC_50) then
		
			-- get freeze
			if (chipselect = '1' and write = '1' and address = "11001") then
				freeze_k <= writedata(0);
			end if;
		
			-- get full_screen
			if (chipselect = '1' and write = '1' and address = "11010") then
				full_screen_k <= writedata(0);
			end if;
		
			-- get number of stored images
			if (chipselect = '1' and write = '1' and address = "01100") then
				num_images <= unsigned(writedata(9 downto 0));
			end if;
			
			-- delete image
			if (chipselect = '1' and read = '1' and address = "01101") then
				if (del_image = '1') then
					readdata <= x"0001";
				else
					readdata <= x"0000";
				end if;
			end if;
			
			-- Send image to the sdram

			-- get x_pos
			if (save_image = '1' and chipselect = '1' and write = '1' and address = "00000") then
				save_image_x <= unsigned(writedata(9 downto 0));
				got_x <= '1';
			end if;
			-- get y_pos
			if (save_image = '1' and chipselect = '1' and write = '1' and address = "00001") then
			  save_image_y <= unsigned(writedata(9 downto 0));
			  got_y <= '1';
			end if;
			-- get request
			if (chipselect = '1' and write = '1' and address = "00010") then
				-- save_image just went high
				if (save_image = '0' and writedata = x"0001") then
					save_image <= '1';
					got_x <= '0';
					got_y <= '0';
					sent_data <= '0';
				-- save_image just went low
				elsif (save_image = '1' and writedata = x"0000") then
					save_image <= '0';
				end if;
			end if;
			-- send data and set ack to 1
			if (save_image = '1' and got_x = '1' and got_y = '1' and chipselect = '1' and read = '1' and address = "00011") then
				readdata <= std_logic_vector(sram_data);
				sent_data <= '1';
				save_image <= '0';
			end if;
			-- tell nios to send request so that i can send data
			if chipselect = '1' and read = '1' and address = "00100" then
				if send_image = '1' and save_image = '0' then
						readdata <= x"0001";
				else
						readdata <= x"0000";
				end if;
			end if;
	
			-- Get an image from the sdram
	
			-- get x_pos
			if (getting_image = '1' and chipselect = '1' and write = '1' and address = "00101") then
				get_image_x <= unsigned(writedata(9 downto 0));
				got_x2 <= '1';
			end if;
			-- get y_pos
			if (getting_image = '1' and chipselect = '1' and write = '1' and address = "00110") then
			  get_image_y <= unsigned(writedata(9 downto 0));
			  got_y2 <= '1';
			end if;
			-- get pixel
			if (getting_image = '1' and chipselect = '1' and write = '1' and address = "00111") then
			  pixel <= unsigned(writedata);
			  got_data <= '1';
			end if;
			-- get request
			if (chipselect = '1' and write = '1' and address = "01000") then
				-- getting_image just went high
				if (getting_image = '0' and writedata = x"0001") then
					getting_image <= '1';
					got_x2 <= '0';
					got_y2 <= '0';
					got_data <= '0';
				-- getting_image just went low
				elsif (got_data = '1' and writedata = x"0000") then
					getting_image <= '0';
					got_data <= '0';
				end if;
			end if;
			-- send data and set ack to 1
			if (getting_image = '1' and got_x2 = '1' and got_y2 = '1' and chipselect = '1' and read = '1' and address = "01001") then
				readdata <= std_logic_vector(pixel);
				got_data <= '1';
				getting_image <= '0';
			end if;
			-- tell nios to send request so that i can send data
			if chipselect = '1' and read = '1' and address = "01010" then
				if getting_image = '0' then
					if move_left = '1' then
						readdata <= x"0001";
					elsif move_right = '1' then
					  readdata <= x"0002";
					end if;
				else
						readdata <= x"0000";
				end if;
			end if;
			
			-- Get thumbnails
			
		end if;
	end process;
	
	process (OSC_27)
	variable count : integer := 0;
	begin
		if rising_edge(OSC_50) then
			count := count + 1;
			if count = 9_000_000 then
				count := 0;
				if move_right = '1' and current_image < num_images - 1 then
					current_image <= current_image + 1;
				elsif move_left = '1' and current_image > 0 then
					current_image <= current_image - 1;
				end if;
			end if;
		end if;
	end process;
				

	-- Determine where within the vga screen
	is_within_screen <= '1' when (vga_x_count >= screen_x_offset and vga_x_count <= screen_x_size + screen_x_offset)
											and (vga_y_count >= screen_y_offset and vga_y_count <= screen_y_size + screen_y_offset) else '0';
	is_within_image <= '1' when (screen_x_count >= image_x_offset and screen_x_count <= image_x_size + image_x_offset)
											and (screen_y_count >= image_y_offset and screen_y_count <= image_y_size + image_y_offset) else '0';
	is_image_frame <= '1' when not is_within_image = '1' and (screen_x_count >= image_x_offset - image_border and screen_x_count <= image_x_size + image_x_offset + image_border)
											and (screen_y_count >= image_y_offset - image_border and screen_y_count <= image_y_size + image_y_offset + image_border) else '0';
	is_within_image_list <= '1' when is_within_screen = '1' and screen_y_count > (image_y_offset + image_y_size + image_border) else '0';
	
	-- Viewable screen
	screen_x_count <= vga_x_count - screen_x_offset when vga_x_count >= screen_x_offset else (others => '0');
	screen_y_count <= vga_y_count - screen_y_offset when vga_y_count >= screen_y_offset else (others => '0');
	
	-- Image section
	image_x_count <= screen_x_count - image_x_offset when screen_x_count >= image_x_offset;
	image_y_count <= screen_y_count - image_y_offset when screen_y_count >= image_y_offset;
	
	-- Image list section
	image_list_x_count <= screen_x_count;
	image_list_y_count <= screen_y_count - (image_y_offset + image_y_size + image_border);
	
	
	-- SRAM
	
	-- Addresses
	screen_addr <= unsigned(screen_x_count(9 downto 1)) + unsigned(screen_y_count(9 downto 1)) * image_x_size;
	image_addr <= unsigned(image_x_count(8 downto 0)) + unsigned(image_y_count(8 downto 0)) * image_x_size;
	--save_image_addr <= unsigned(save_image_x(9 downto 1)) + unsigned(save_image_y(9 downto 1)) * (screen_x_size / 2);
	save_image_addr <= unsigned(save_image_x(8 downto 0)) + unsigned(save_image_y(8 downto 0)) * image_x_size;
	get_image_addr <= unsigned(get_image_x(8 downto 0)) + unsigned(get_image_y(8 downto 0)) * image_x_size;
	ram_addr <= get_image_addr when (got_data = '1') else save_image_addr when (save_image = '1') else screen_addr when (full_screen = '1' or freeze = '0') else image_addr;
	-- Write data
	write_to_ram <= got_data or (not freeze and is_within_screen and not save_image and screen_y_count(0) and screen_x_count(0));
	write_data <= std_logic_vector(pixel) when got_data = '1' else red_r(9 downto 5) & green_r(9 downto 4) & blue_r(9 downto 5);
	-- Read data
	req_data <= sram_data;
	red_pixel <= sram_data(15 downto 11) & "00000";
	green_pixel <= sram_data(10 downto 5) & "0000";
	blue_pixel <= sram_data(4 downto 0) & "00000";
	sram_frame_buffer: sram port map(
		chipselect	=> '1',
		write				=> write_to_ram,
		read				=> not got_data and (freeze or save_image),
		address			=> std_logic_vector(ram_addr),
		writedata		=> write_data,
		byteenable	=> "11",
		readdata		=> sram_data,
		SRAM_DQ			=> SRAM_DQ,
		SRAM_ADDR		=> SRAM_ADDR,
		SRAM_UB_N		=> SRAM_UB_N,
		SRAM_LB_N		=> SRAM_LB_N,
		SRAM_WE_N		=> SRAM_WE_N,
		SRAM_CE_N		=> SRAM_CE_N,
		SRAM_OE_N		=> SRAM_OE_N);
	

	-- VGA display

--	VGA_R <= red_r when freeze = '0' or save_image = '1' else
--					 sram_data(15 downto 11) & "00000" when is_within_screen = '1' else
--					 (others => '0');
--	VGA_G <= green_r when freeze = '0' or save_image = '1' else
--					 sram_data(10 downto 5) & "0000" when is_within_screen = '1' else
--					 (others => '0');
--	VGA_B <= blue_r when freeze = '0' or save_image = '1' else
--					 sram_data(4 downto 0) & "00000" when is_within_screen = '1' else
--					 (others => '0');
					

	
	process (OSC_27)
	variable vga_r_r, vga_g_r, vga_b_r : std_logic_vector(9 downto 0);
	begin
		if rising_edge(OSC_27) then
					-- when sending data to the sram
					if ((got_data = '1' or save_image = '1') and (freeze = '1' and full_screen = '0')) then
							if is_within_image = '1' then
								VGA_R <= red_pixel;
								VGA_G <= green_pixel;
								VGA_B <= blue_pixel;
							elsif is_image_frame = '1' then
								VGA_R <= (others => '1');
								VGA_G <= (others => '1');
								VGA_B <= (others => '1');
							else
								VGA_R <= (others => '0');
								VGA_G <= (others => '0');
								VGA_B <= (others => '0');
							end if;
					elsif (freeze = '0' and full_screen = '1' and is_within_screen = '1') or
									(freeze = '0' and full_screen = '0' and is_within_image = '1') then
								VGA_R <= red_r;
								VGA_G <= green_r;
								VGA_B <= blue_r;
					elsif (freeze = '1' and full_screen = '1' and is_within_screen = '1') or
									(freeze = '1' and full_screen = '0' and is_within_image = '1') then
							VGA_R <= red_pixel;
							VGA_G <= green_pixel;
							VGA_B <= blue_pixel;
					elsif full_screen = '0' and is_image_frame = '1' then
							VGA_R <= (others => '1');
							VGA_G <= (others => '1');
							VGA_B <= (others => '1');
					else
							VGA_R <= (others => '0');
							VGA_G <= (others => '0');
							VGA_B <= (others => '0');
					end if;
					
		end if;
	end process;
end rtl;
