library ieee;
  use ieee.std_logic_1164.all;
  use ieee.numeric_std.all;
  
entity JpegTop is
  port (
    signal CLOCK_50 : in std_logic;                -- 50 MHz clock
	signal LEDR : out std_logic_vector(17 downto 0);      -- Red LEDs
	signal SW : in std_logic_vector(17 downto 0);         -- DPDT switches

    SRAM_DQ : inout std_logic_vector(15 downto 0); -- Data bus 16 Bits
    SRAM_ADDR : out std_logic_vector(17 downto 0); -- Address bus 18 Bits
    SRAM_UB_N,                                     -- High-byte Data Mask
    SRAM_LB_N,                                     -- Low-byte Data Mask
    SRAM_WE_N,                                     -- Write Enable
    SRAM_CE_N,                                     -- Chip Enable
    SRAM_OE_N : out std_logic                      -- Output Enable
    );
	
end JpegTop;

architecture rtl of JpegTop is
  --OPB exports
  signal OPB_ABus          : std_logic_vector(31 downto 0);
  signal OPB_BE            : std_logic_vector(3 downto 0);
  signal OPB_DBus_in       : std_logic_vector(31 downto 0);
  signal OPB_RNW           : std_logic;
  signal OPB_select        : std_logic;
  signal OPB_DBus_out      : std_logic_vector(31 downto 0);
  signal OPB_XferAck       : std_logic;
  signal OPB_retry         : std_logic;
  signal OPB_toutSup       : std_logic;
  signal OPB_errAck        : std_logic;

  --pixel line buffer
  signal iram_wdata			: std_logic_vector(23 downto 0);
  signal iram_wren 			: std_logic;
  signal iram_fifo_afull 	: std_logic;

  --output ram
  signal ram_byte 			: std_logic_vector(7 downto 0);
  signal ram_wren 			: std_logic;
  signal ram_wraddr 		: std_logic_vector(23 downto 0);
  signal outif_almost_full 	: std_logic := '0';
  
  signal outram_raddress	: std_logic_vector(13 downto 0);		--outram 14 bits
  signal outram_readdata	: std_logic_vector(7 downto 0);

  signal reset_n 	: std_logic :=	'1';
  signal RST 		: std_logic := 	'0';

--  signal start_test : std_logic :=	'0';
  
begin

	RST <= (not reset_n) or SW(0);
	LEDR(0) <= RST;
    LEDR(17 downto 1) <= (others => '0');

  nios : entity work.nios_system port map (
    clk                          => CLOCK_50,
    reset_n                      => reset_n,

    SRAM_ADDR_from_the_sram      => SRAM_ADDR,
    SRAM_CE_N_from_the_sram      => SRAM_CE_N,
    SRAM_DQ_to_and_from_the_sram => SRAM_DQ,
    SRAM_LB_N_from_the_sram      => SRAM_LB_N,
    SRAM_OE_N_from_the_sram      => SRAM_OE_N,
    SRAM_UB_N_from_the_sram      => SRAM_UB_N,
    SRAM_WE_N_from_the_sram      => SRAM_WE_N,

    raddr_from_the_readram	   	 	  => outram_raddress,
    q_to_the_readram			 	  => outram_readdata,

    OPB_DBus_out_to_the_avalontrans	  => OPB_DBus_out,
    OPB_XferAck_to_the_avalontrans	  => OPB_XferAck,
    OPB_retry_to_the_avalontrans	  => OPB_retry,
    OPB_toutSup_to_the_avalontrans	  => OPB_toutSup,
    OPB_errAck_to_the_avalontrans 	  => OPB_errAck,
    OPB_ABus_from_the_avalontrans 	  => OPB_ABus,
    OPB_BE_from_the_avalontrans 	  => OPB_BE,
    OPB_DBus_in_from_the_avalontrans  => OPB_DBus_in,
    OPB_RNW_from_the_avalontrans 	  => OPB_RNW,
    OPB_select_from_the_avalontrans   => OPB_select,

    iram_wdata_from_the_savepixel	  => iram_wdata,
    iram_wren_from_the_savepixel	  => iram_wren,
    fifo_almost_full_to_the_savepixel => iram_fifo_afull

    );

  jpeg : entity work.JpegEnc port map (
        CLK                => CLOCK_50,
        RST                => RST,

        -- OPB
        OPB_ABus           => OPB_ABus,
        OPB_BE             => OPB_BE,
        OPB_DBus_in        => OPB_DBus_in,
        OPB_RNW            => OPB_RNW,
        OPB_select         => OPB_select,
        OPB_DBus_out       => OPB_DBus_out,
        OPB_XferAck        => OPB_XferAck,
        OPB_retry          => OPB_retry,
        OPB_toutSup        => OPB_toutSup,
        OPB_errAck         => OPB_errAck,  

        -- IMAGE RAM
        iram_wdata         => iram_wdata,
        iram_wren          => iram_wren,
        iram_fifo_afull    => iram_fifo_afull,

        -- OUT RAM
        ram_byte           => ram_byte,
        ram_wren           => ram_wren,  
        ram_wraddr         => ram_wraddr,
        outif_almost_full  => '0' --outif_almost_full
    );

  outram : entity work.OutputRAM port map (
		clk					=> CLOCK_50,
		raddr				=> outram_raddress,
		waddr				=> ram_wraddr(13 downto 0),		--outram 14 bits
		data				=> ram_byte,
		we					=> ram_wren,
		q					=> outram_readdata
	);

end rtl;
