--
-- Video Chat System top-level module
--
-- Modified from DE2_TOP
-- by Stephen A. Edwards, Columbia University, sedwards@cs.columbia.edu
--
-- From an original by Terasic Technology, Inc.
-- (DE2_TOP.v, part of the DE2 system board CD supplied by Altera)
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity embedded_image_capture is

  port (
    -- Clocks
    
    OSC_27,                                      	 -- 27 MHz
    OSC_50,                                      	 -- 50 MHz
    EXT_CLOCK : in std_logic;                      -- External Clock

    -- Buttons and switches
    
    KEY : in std_logic_vector(3 downto 0);         -- Push buttons
    DPDT_SW : in std_logic_vector(17 downto 0);		 -- DPDT switches

    -- LED displays
	LED_RED : out std_logic_vector(17 downto 0);      -- Red LEDs

    HEX0, HEX1, HEX2, HEX3, HEX4, HEX5, HEX6, HEX7 -- 7-segment displays
       : out std_logic_vector(6 downto 0);

    -- SRAM
    
    SRAM_DQ : inout std_logic_vector(15 downto 0); -- Data bus 16 Bits
    SRAM_ADDR : out std_logic_vector(17 downto 0); -- Address bus 18 Bits
    SRAM_UB_N,                                     -- High-byte Data Mask 
    SRAM_LB_N,                                     -- Low-byte Data Mask 
    SRAM_WE_N,                                     -- Write Enable
    SRAM_CE_N,                                     -- Chip Enable
    SRAM_OE_N : out std_logic;                     -- Output Enable

    -- SDRAM
   
    DRAM_DQ : inout std_logic_vector(15 downto 0); -- Data Bus
    DRAM_ADDR : out std_logic_vector(11 downto 0); -- Address Bus    
    DRAM_LDQM,                                     -- Low-byte Data Mask 
    DRAM_UDQM,                                     -- High-byte Data Mask
    DRAM_WE_N,                                     -- Write Enable
    DRAM_CAS_N,                                    -- Column Address Strobe
    DRAM_RAS_N,                                    -- Row Address Strobe
    DRAM_CS_N,                                     -- Chip Select
    DRAM_BA_0,                                     -- Bank Address 0
    DRAM_BA_1,                                     -- Bank Address 1
    DRAM_CLK,                                      -- Clock
    DRAM_CKE : out std_logic;                      -- Clock Enable

    -- I2C bus
    
    I2C_DATA : inout std_logic; -- I2C Data
    I2C_CLK : inout std_logic;   -- I2C Clock

    -- VGA output
    
    VGA_CLK,                                    -- Clock
    VGA_HS,                                     -- H_SYNC
    VGA_VS,                                     -- V_SYNC
    VGA_BLANK,                                  -- BLANK
    VGA_SYNC : out std_logic;                   -- SYNC
    VGA_R,                                      -- Red[9:0]
    VGA_G,                                      -- Green[9:0]
    VGA_B : out std_logic_vector(9 downto 0);   -- Blue[9:0]

    -- Video Decoder
    
    TD_DATA : in std_logic_vector(7 downto 0);  -- Data bus 8 bits
    TD_HS,                                      -- H_SYNC
    TD_VS : in std_logic;                       -- V_SYNC
    TD_RESET : out std_logic;                   -- Reset
    
    --  Ethernet Interface
    
    ENET_DATA : inout unsigned(15 downto 0);    -- DATA bus 16 Bits
    ENET_CMD,           -- Command/Data Select, 0 = Command, 1 = Data
    ENET_CS_N,                                  -- Chip Select
    ENET_WR_N,                                  -- Write
    ENET_RD_N,                                  -- Read
    ENET_RST_N,                                 -- Reset
    ENET_CLK : out std_logic;                   -- Clock 25 MHz
    ENET_INT : in std_logic;                    -- Interrupt

    -- RS-232 interface

    UART_TXD : out std_logic;                      -- UART transmitter   
    UART_RXD : in std_logic;                       -- UART receiver

    -- 16 X 2 LCD Module
    
    LCD_ON,                     -- Power ON/OFF
    LCD_BLON,                   -- Back Light ON/OFF
    LCD_RW,                     -- Read/Write Select, 0 = Write, 1 = Read
    LCD_EN,                     -- Enable
    LCD_RS : out std_logic;     -- Command/Data Select, 0 = Command, 1 = Data
    LCD_DATA : inout std_logic_vector(7 downto 0); -- Data bus 8 bits

    -- PS/2 port

    PS2_DATA,                   -- Data
    PS2_CLK : inout std_logic    -- Clock

    );
	signal rst, freeze, full_screen, send_image, del_image: std_logic;
	signal move_right, move_left: std_logic;
	signal sdram_clk	: std_logic;
	signal mSEG7_HEX : std_logic_vector(31 downto 0);
	signal DATA_to_OTG : std_logic_vector(15 downto 0);
	signal DATA_to_SRAM : std_logic_vector(15 downto 0);
	--	I2C
	signal SCL_OE_N, SDA_OE_N, SCL, SDA : std_logic;

	--	CPU
	signal CPU_RESET : std_logic;
	signal DRAM_BA : std_logic_vector(1 downto 0);
	signal DRAM_QM : std_logic_vector(1 downto 0);
	
end embedded_image_capture;


architecture datapath of embedded_image_capture is

------------------------------------------------------------------
------------------ J P E G   S I G N A L S -----------------------
------------------------------------------------------------------

  --OPB exports
  signal OPB_ABus          : std_logic_vector(31 downto 0);
  signal OPB_BE            : std_logic_vector(3 downto 0);
  signal OPB_DBus_in       : std_logic_vector(31 downto 0);
  signal OPB_RNW           : std_logic;
  signal OPB_select        : std_logic;
  signal OPB_DBus_out      : std_logic_vector(31 downto 0);
  signal OPB_XferAck       : std_logic;
  signal OPB_retry         : std_logic;
  signal OPB_toutSup       : std_logic;
  signal OPB_errAck        : std_logic;

  --pixel line buffer
  signal iram_wdata			: std_logic_vector(23 downto 0);
  signal iram_wren 			: std_logic;
  signal iram_fifo_afull 	: std_logic;

  --output ram
  signal ram_byte 			: std_logic_vector(7 downto 0);
  signal ram_wren 			: std_logic;
  signal ram_wraddr 		: std_logic_vector(23 downto 0);
  signal outif_almost_full 	: std_logic := '0';
  
  signal outram_raddress	: std_logic_vector(13 downto 0);		--outram 14 bits
  signal outram_readdata	: std_logic_vector(7 downto 0);
  signal jpeg_reset			: std_logic := '0';

------------------------------------------------------------------
------------- E N D  J P E G  S I G N A L S ----------------------
------------------------------------------------------------------


	component I2C_AV_Config port(
		-- Host Side
		iCLK,
		iRST_N		: in std_logic;
		-- I2C Side
		I2C_SCLK	: out std_logic;
		I2C_SDAT	: inout std_logic);
	end component;
	
	component Reset_Delay port(
		iCLK : in std_logic;
		oRESET : out std_logic
		);
	end component;
		
	component nios_0 port(	
			-- 1 global signals:
		clk,
		reset_n : std_logic;
		
		------------- J P E G  S T U F F --------------
		
		raddr_from_the_readram 				: out std_logic_vector(13 downto 0);
		q_to_the_readram 					: in std_logic_vector(7 downto 0);

		OPB_DBus_out_to_the_avalontrans 	: in std_logic_vector(31 downto 0);
		OPB_XferAck_to_the_avalontrans 		: in std_logic;
		OPB_retry_to_the_avalontrans 		: in std_logic;
		OPB_toutSup_to_the_avalontrans 		: in std_logic;
		OPB_errAck_to_the_avalontrans 		: in std_logic;
		OPB_ABus_from_the_avalontrans 		: out std_logic_vector(31 downto 0);
		OPB_BE_from_the_avalontrans 		: out std_logic_vector(3 downto 0);
		OPB_DBus_in_from_the_avalontrans	: out std_logic_vector(31 downto 0);
		OPB_RNW_from_the_avalontrans 		: out std_logic;
		OPB_select_from_the_avalontrans 	: out std_logic;
		iram_wdata_from_the_savepixel 		: out std_logic_vector(23 downto 0);
		iram_wren_from_the_savepixel 		: out std_logic;
		fifo_almost_full_to_the_savepixel 	: in std_logic;		
		
		
		----------- E N D  J P E G  S T U F F ---------
		

		-- the_TV_to_VGA_inst
		OSC_27_to_the_TV_to_VGA_inst,
		OSC_50_to_the_TV_to_VGA_inst : in  std_logic;
		SRAM_ADDR_from_the_TV_to_VGA_inst : out std_logic_vector(17 downto 0);
		SRAM_CE_N_from_the_TV_to_VGA_inst : out std_logic;
		SRAM_DQ_to_and_from_the_TV_to_VGA_inst  : inout std_logic_vector(15 downto 0);
		SRAM_LB_N_from_the_TV_to_VGA_inst,
		SRAM_OE_N_from_the_TV_to_VGA_inst,
		SRAM_UB_N_from_the_TV_to_VGA_inst,
		SRAM_WE_N_from_the_TV_to_VGA_inst : out std_logic;
		TD_D_to_the_TV_to_VGA_inst : in std_logic_vector(7 downto 0);
		TD_HS_to_the_TV_to_VGA_inst,
		TD_VS_to_the_TV_to_VGA_inst : in  std_logic;
		VGA_SYNC_from_the_TV_to_VGA_inst,
		VGA_VS_from_the_TV_to_VGA_inst,
		VGA_BLANK_from_the_TV_to_VGA_inst,
		VGA_HS_from_the_TV_to_VGA_inst,
		VGA_CLOCK_from_the_TV_to_VGA_inst : out  std_logic;
		VGA_G_from_the_TV_to_VGA_inst,
		VGA_B_from_the_TV_to_VGA_inst,
		VGA_R_from_the_TV_to_VGA_inst : out std_logic_vector(9 downto 0);
		reset_to_the_TV_to_VGA_inst : in std_logic;
		
		freeze_b_to_the_TV_to_VGA_inst,
		full_screen_b_to_the_TV_to_VGA_inst,
		send_image_to_the_TV_to_VGA_inst,
		del_image_to_the_TV_to_VGA_inst,
		move_right_to_the_TV_to_VGA_inst,
		move_left_to_the_TV_to_VGA_inst : in std_logic;

		-- the_sdram
		zs_addr_from_the_sdram_0 : out std_logic_vector(11 downto 0);
		zs_dq_to_and_from_the_sdram_0 : inout std_logic_vector(15 downto 0);
		zs_dqm_from_the_sdram_0,
		zs_ba_from_the_sdram_0   : out std_logic_vector(1 downto 0);
		zs_cas_n_from_the_sdram_0,
		zs_cke_from_the_sdram_0,
		zs_cs_n_from_the_sdram_0,
		zs_ras_n_from_the_sdram_0,
		zs_we_n_from_the_sdram_0  : out std_logic;
		

		-- the_I2C_0
		scl_pad_i_to_the_I2C_0 : in std_logic;
		scl_pad_o_from_the_I2C_0,
		scl_padoen_o_from_the_I2C_0  : out std_logic;
		sda_pad_i_to_the_I2C_0 : in std_logic;
		sda_pad_o_from_the_I2C_0,
		sda_padoen_o_from_the_I2C_0,
		
		-- the_lcd_16207_0
		LCD_E_from_the_lcd_16207_0,
		LCD_RS_from_the_lcd_16207_0,
		LCD_RW_from_the_lcd_16207_0 : out std_logic;
		LCD_data_to_and_from_the_lcd_16207_0 : inout std_logic_vector(7 downto 0);

		-- the_SEG7_Display
		oSEG0_from_the_SEG7_Display,
		oSEG1_from_the_SEG7_Display,
		oSEG2_from_the_SEG7_Display,
		oSEG3_from_the_SEG7_Display,
		oSEG4_from_the_SEG7_Display,
		oSEG5_from_the_SEG7_Display,
		oSEG6_from_the_SEG7_Display,
		oSEG7_from_the_SEG7_Display : out std_logic_vector(6 downto 0);


		-- the_DM9000A
		ENET_DATA_to_and_from_the_DM9000A : inout unsigned(15 downto 0);
		ENET_CMD_from_the_DM9000A,
		ENET_CS_N_from_the_DM9000A,
		ENET_RD_N_from_the_DM9000A,
		ENET_RST_N_from_the_DM9000A,
		ENET_WR_N_from_the_DM9000A : out std_logic;
		ENET_INT_to_the_DM9000A : in std_logic;

		-- the_uart_0
		rxd_to_the_uart_0 : in std_logic;
		txd_from_the_uart_0 : out std_logic;

		-- the_ps2_0
		PS2_CLK_to_and_from_the_ps2_0,
		PS2_DAT_to_and_from_the_ps2_0  : inout std_logic
		);
	end component;
	
begin

	i2c: I2C_AV_Config port map(
		--	Host Side
		iCLK		=> OSC_27,
		iRST_N		=> rst,
		--	I2C Side
		I2C_SCLK	=> I2C_CLK,
		I2C_SDAT	=> I2C_DATA);



	nios: nios_0 port map (
		-- 1 global signals:
		clk		=> OSC_50,
		reset_n	=> CPU_RESET,

		-- the_TV_to_VGA_inst
		OSC_27_to_the_TV_to_VGA_inst 			=> OSC_27,
		OSC_50_to_the_TV_to_VGA_inst 			=> OSC_50,
		SRAM_ADDR_from_the_TV_to_VGA_inst	=> SRAM_ADDR,
		SRAM_CE_N_from_the_TV_to_VGA_inst	=> SRAM_CE_N,
		SRAM_DQ_to_and_from_the_TV_to_VGA_inst => SRAM_DQ,
		SRAM_LB_N_from_the_TV_to_VGA_inst	=> SRAM_LB_N,
		SRAM_OE_N_from_the_TV_to_VGA_inst	=> SRAM_OE_N,
		SRAM_UB_N_from_the_TV_to_VGA_inst	=> SRAM_UB_N,
		SRAM_WE_N_from_the_TV_to_VGA_inst	=> SRAM_WE_N,
		TD_D_to_the_TV_to_VGA_inst			=> TD_DATA,
		TD_HS_to_the_TV_to_VGA_inst			=> TD_HS,
		TD_VS_to_the_TV_to_VGA_inst			=> TD_VS,
		VGA_BLANK_from_the_TV_to_VGA_inst	=> VGA_BLANK,
		VGA_B_from_the_TV_to_VGA_inst		=> VGA_B,
		VGA_CLOCK_from_the_TV_to_VGA_inst	=> VGA_clk,
		VGA_G_from_the_TV_to_VGA_inst		=> VGA_G,
		VGA_HS_from_the_TV_to_VGA_inst		=> VGA_HS,
		VGA_R_from_the_TV_to_VGA_inst		=> VGA_R,
		VGA_SYNC_from_the_TV_to_VGA_inst	=> VGA_SYNC,
		VGA_VS_from_the_TV_to_VGA_inst		=> VGA_VS,
		freeze_b_to_the_TV_to_VGA_inst		=> freeze,
		full_screen_b_to_the_TV_to_VGA_inst		=> full_screen,
		send_image_to_the_TV_to_VGA_inst		=> send_image,
		del_image_to_the_TV_to_VGA_inst		=> del_image,
		move_right_to_the_TV_to_VGA_inst		=> move_right,
		move_left_to_the_TV_to_VGA_inst		=> move_left,
		reset_to_the_TV_to_VGA_inst			=> rst,

		-- the_sdram
		zs_addr_from_the_sdram_0			=> DRAM_ADDR,
		zs_ba_from_the_sdram_0			=> DRAM_BA,
		zs_cas_n_from_the_sdram_0			=> DRAM_CAS_N,
		zs_cke_from_the_sdram_0			=> DRAM_CKE,
		zs_cs_n_from_the_sdram_0			=> DRAM_CS_N,
		zs_dq_to_and_from_the_sdram_0		=> DRAM_DQ,
		zs_dqm_from_the_sdram_0			=> DRAM_QM,
		zs_ras_n_from_the_sdram_0			=> DRAM_RAS_N,
		zs_we_n_from_the_sdram_0			=> DRAM_WE_N,
		

		-- the_I2C_0
		scl_pad_i_to_the_I2C_0	=>	I2C_CLK,
		scl_pad_o_from_the_I2C_0	=>	SCL,
		scl_padoen_o_from_the_I2C_0	=>	SCL_OE_N,
		sda_pad_i_to_the_I2C_0	=>	I2C_DATA,
		sda_pad_o_from_the_I2C_0	=>	SDA,
		sda_padoen_o_from_the_I2C_0	=>	SDA_OE_N,
		
		-- the_lcd_16207_0
		LCD_E_from_the_lcd_16207_0	=>	LCD_EN,
		LCD_RS_from_the_lcd_16207_0	=>	LCD_RS,
		LCD_RW_from_the_lcd_16207_0	=>	LCD_RW,
		LCD_data_to_and_from_the_lcd_16207_0	=>	LCD_DATA,

		-- the_SEG7_Display
		oSEG0_from_the_SEG7_Display	=>	HEX0,
		oSEG1_from_the_SEG7_Display	=>	HEX1,
		oSEG2_from_the_SEG7_Display	=>	HEX2,
		oSEG3_from_the_SEG7_Display	=>	HEX3,
		oSEG4_from_the_SEG7_Display	=>	HEX4,
		oSEG5_from_the_SEG7_Display	=>	HEX5,
		oSEG6_from_the_SEG7_Display	=>	HEX6,
		oSEG7_from_the_SEG7_Display	=>	HEX7,


		-- the_DM9000A
		ENET_CMD_from_the_DM9000A	=>	ENET_CMD,
		ENET_CS_N_from_the_DM9000A	=>	ENET_CS_N,
		ENET_DATA_to_and_from_the_DM9000A	=>	ENET_DATA,
		ENET_INT_to_the_DM9000A	=>	ENET_INT,
		ENET_RD_N_from_the_DM9000A	=>	ENET_RD_N,
		ENET_RST_N_from_the_DM9000A	=>	ENET_RST_N,
		ENET_WR_N_from_the_DM9000A	=>	ENET_WR_N,

		-- the_uart_0
		rxd_to_the_uart_0	=>	UART_RXD,
		txd_from_the_uart_0	=>	UART_TXD,

		-- the_ps2_0
		PS2_CLK_to_and_from_the_ps2_0	=>	PS2_CLK,
		PS2_DAT_to_and_from_the_ps2_0	=>	PS2_DATA,
		
		
		------------------- J P E G  P E R I P H E R A L S -------------------
		
		raddr_from_the_readram	   	 	  => outram_raddress,
		q_to_the_readram			 	  => outram_readdata,

		OPB_DBus_out_to_the_avalontrans	  => OPB_DBus_out,
		OPB_XferAck_to_the_avalontrans	  => OPB_XferAck,
		OPB_retry_to_the_avalontrans	  => OPB_retry,
		OPB_toutSup_to_the_avalontrans	  => OPB_toutSup,
		OPB_errAck_to_the_avalontrans 	  => OPB_errAck,
		OPB_ABus_from_the_avalontrans 	  => OPB_ABus,
		OPB_BE_from_the_avalontrans 	  => OPB_BE,
		OPB_DBus_in_from_the_avalontrans  => OPB_DBus_in,
		OPB_RNW_from_the_avalontrans 	  => OPB_RNW,
		OPB_select_from_the_avalontrans   => OPB_select,

		iram_wdata_from_the_savepixel	  => iram_wdata,
		iram_wren_from_the_savepixel	  => iram_wren,
		fifo_almost_full_to_the_savepixel => iram_fifo_afull
		
		----------------- E N D  J P E G  P E R I P H E R A L S --------------
			
	);
	
	------------------------ J P E G  M O D U L E -------------------
	
	jpeg_reset <= DPDT_SW(1);
	LED_RED(1) <= jpeg_reset;
	
	  jpeg : entity work.JpegEnc port map (
        CLK                => OSC_50,
        RST                => jpeg_reset,

        -- OPB
        OPB_ABus           => OPB_ABus,
        OPB_BE             => OPB_BE,
        OPB_DBus_in        => OPB_DBus_in,
        OPB_RNW            => OPB_RNW,
        OPB_select         => OPB_select,
        OPB_DBus_out       => OPB_DBus_out,
        OPB_XferAck        => OPB_XferAck,
        OPB_retry          => OPB_retry,
        OPB_toutSup        => OPB_toutSup,
        OPB_errAck         => OPB_errAck,  

        -- IMAGE RAM
        iram_wdata         => iram_wdata,
        iram_wren          => iram_wren,
        iram_fifo_afull    => iram_fifo_afull,

        -- OUT RAM
        ram_byte           => ram_byte,
        ram_wren           => ram_wren,  
        ram_wraddr         => ram_wraddr,
        outif_almost_full  => '0' --outif_almost_full
    );

	-------------- E N D  J P E G  M O D U L E --------------
	
	----------------- O U T P U T  M E R O R Y ---------------

	outram : entity work.OutputRAM port map (
		clk					=> OSC_50,
		raddr				=> outram_raddress,
		waddr				=> ram_wraddr(13 downto 0),		--outram 14 bits
		data				=> ram_byte,
		we					=> ram_wren,
		q					=> outram_readdata
	);

	----------------- E N D  O U T P U T  R A M -----------------------
	
	
	
	sdram_pll: work.sdram_pll port map (
		inclk0	=> OSC_50,
		c0			=> DRAM_CLK 
	);
	
	delay: Reset_Delay port map (
		iCLK => OSC_50,
		oRESET => CPU_RESET
		);
	
--	DRAM_CLK  <= sdram_clk;
--	DRAM_UDQM <= DRAM_QM(1);
--	DRAM_LDQM <= DRAM_QM(0);
--	DRAM_BA_1 <= DRAM_BA(1);
--	DRAM_BA_0 <= DRAM_BA(0);

	-- TV
	TD_RESET	<= '1';	--	Bypress 27 MHz
	rst			<= DPDT_SW(0);

	--	I2C Tri-inout
--	I2C_CLK	<=	'Z' when SCL_OE_N	= '1' else SCL;
--	I2C_DATA	<=	'Z' when SDA_OE_N = '1' else SDA;
	
  DRAM_BA_1 <= DRAM_BA(1);
  DRAM_BA_0 <= DRAM_BA(0);
  DRAM_UDQM <= DRAM_QM(1);
  DRAM_LDQM <= DRAM_QM(0);

	process (OSC_50)
	variable clk	: std_logic := '0';
	begin
		if rising_edge(OSC_50) then
			clk := not clk;
			ENET_CLK <= clk;
		end if;
	end process;

	-- Handle asynchrounous signals
	process (OSC_27)
	variable count : integer := 0;
	begin
		if rising_edge(OSC_27) then
			count := count + 1;
			if count = 2_700_000 then
				count := 0;
				freeze <= DPDT_SW(17);
				full_screen <= not DPDT_SW(16);
				send_image <= not key(3);
				del_image <= not key(2) and key(3);
				move_left <= not key(1);
				move_right <= not key(0) and key(1);
			end if;
		end if;
	end process;
			
	
	

	
end datapath;
