#include <string.h>
#include <stdio.h>
#include <alt_types.h>
#include "DM9000A.h"
#include "network.h"
#include "misc.h"
#include "basic_io.h"

const unsigned char SOURCE_MAC_ADDR[] = { 0x01, 0x60, 0x6E, 0x11, 0x02, 0x41 };
//const unsigned char DEST_MAC_ADDR[] = { 0x00, 0x1A, 0x80, 0xD6, 0x4D, 0xC1 };
const unsigned char DEST_MAC_ADDR[] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };

const int ETHERNET_HEADER_SIZE = 14;
const int IP_HEADER_SIZE = 20;
const int UDP_HEADER_SIZE = 8;

unsigned int network_interrupt_number;

unsigned int network_rx_buffer_length;
unsigned char network_rx_buffer[1600];

extern unsigned char network_udp_packet[];

// Send a packet of data with the specified size, if over MTU split
// the packet into chunks
void network_send_packet(const char* data, const unsigned int size) {
  static unsigned int bytesSent = 0;
  Packet packet;
  
  int i; 
  for(i=0; i<size; i+=NETWORK_MAX_PACKET_SIZE) {
    // truncate the packet if needed
    unsigned int packetSize = (size-i);
    if(packetSize > NETWORK_MAX_PACKET_SIZE) packetSize = NETWORK_MAX_PACKET_SIZE; 
    network_format_packet(&packet, data+i, packetSize);
                         
    int res = TransmitPacket(packet.data, packet.size);
    if (res == DMFE_SUCCESS) { 
      //printf("Packet sent: %d\n", packet.size);
      bytesSent += size;
      outport(SEG7_DISPLAY_BASE, bytesSent);
    } else {
      printf("\nNetwork Failure: Transmission failed\n"); 
    }
    
    // free memory allocated in network_format_packet
    free(packet.data);
  }
      
} // end network_send_packet

// Format a packet for delivery in a Packet structure.  If size is less 
// than MIN_PACKET_SIZE, add '0' bytes of padding to the packet.  Caller
// must free packet->data. 
void network_format_packet(
    Packet* packet, const char* payload, const unsigned int payloadSize) {
      
  // add extra byte for null termination
  int size = NETWORK_UDP_PACKET_SIZE + payloadSize;
  if(size < MIN_PACKET_SIZE) size = MIN_PACKET_SIZE;
  char* packetData = (char*) malloc(size);
  // clear the packet
  memset(packetData, 0, size);
  memcpy(packetData, network_udp_packet, NETWORK_UDP_PACKET_SIZE);
  
  /* Set MAC addresses */
  memcpy(packetData, DEST_MAC_ADDR, 6);  
  memcpy(packetData + 6, SOURCE_MAC_ADDR, 6);
  //memcpy(packetData, SOURCE_MAC_ADDR, 6);  
  //memcpy(packetData + 6, DEST_MAC_ADDR, 6);
  
  /* Set the UDP Length: Header plus data */
  unsigned short udpLength = UDP_HEADER_SIZE + payloadSize;
  //printf("%10s=%d\n", "UDP LENGTH", udpLength);
  
  packetData[HEADER_OFFSET_UDP_LENGTH_H] = udpLength >> 8;
  packetData[HEADER_OFFSET_UDP_LENGTH_L] = udpLength & 0xFF;
  
  /* Set UDP Payload: copy the message to the udp packet */
  memcpy(packetData + NETWORK_UDP_PACKET_SIZE, payload, payloadSize);   
  
  /* Set the IP Length: Header plus UDP Length */
  unsigned short ipLength = IP_HEADER_SIZE + udpLength;
  //printf("%10s=%d\n", "IP LENGTH", ipLength);
  
  packetData[HEADER_OFFSET_IP_TOTAL_LENGTH_H] = ipLength >> 8;
  packetData[HEADER_OFFSET_IP_TOTAL_LENGTH_L] = ipLength & 0xFF;
    
  /* Set the IP Checksum */
  int ipOffset = ETHERNET_HEADER_SIZE;    
  short ipChecksum = 
    network_calc_ip_checksum(packetData+ipOffset, IP_HEADER_SIZE);

  packetData[HEADER_OFFSET_IP_CHECKSUM_H] = ipChecksum >> 8;
  packetData[HEADER_OFFSET_IP_CHECKSUM_L] = ipChecksum & 0xFF;
         
  char padding = (udpLength%2==1);
  // enable padding if udplength is odd
  //printf("Padding=%d, udpLength=%d\n", padding, udpLength);
  char* udpPayload = (packetData + HEADER_OFFSET_UDP);
  // byte 3 is the MSB of the address  
  char* sourceIp = packetData+HEADER_OFFSET_SOURCE_IP_3;
  char* destIp = packetData+HEADER_OFFSET_DEST_IP_3;  
  
  u16 udpChecksum = 
    network_calc_udp_checksum (udpLength, sourceIp, destIp, padding, udpPayload); 
        
  packetData[HEADER_OFFSET_UDP_CHECKSUM_H] = udpChecksum >> 8;
  packetData[HEADER_OFFSET_UDP_CHECKSUM_L] = udpChecksum & 0xFF;
       
  packet->data = packetData;
  //HEADER_OFFSET_UDP_DATA + payloadSize;
  packet->size = size;
} // end network_format_packet

void network_set_source_ip(unsigned char a3, unsigned char a2, unsigned char a1, unsigned char a0) {
  network_udp_packet[HEADER_OFFSET_SOURCE_IP_3] = a3;
  network_udp_packet[HEADER_OFFSET_SOURCE_IP_2] = a2;
  network_udp_packet[HEADER_OFFSET_SOURCE_IP_1] = a1;
  network_udp_packet[HEADER_OFFSET_SOURCE_IP_0] = a0; 
} // end network_set_source_ip

void network_set_dest_ip(unsigned char a3, unsigned char a2, unsigned char a1, unsigned char a0) {
  network_udp_packet[HEADER_OFFSET_DEST_IP_3] = a3;
  network_udp_packet[HEADER_OFFSET_DEST_IP_2] = a2;
  network_udp_packet[HEADER_OFFSET_DEST_IP_1] = a1;
  network_udp_packet[HEADER_OFFSET_DEST_IP_0] = a0;
} // end network_set_dest_ip

// add up an array of byte values to calculate the checksum
unsigned short network_calc_ip_checksum(unsigned char ipHeader[], const int SIZE) {
  unsigned int checksum = 0;
  int i;
  for(i=0; i<SIZE; i+=2) {
    // even bytes are high order    
    int val = (ipHeader[i] << 8) | ipHeader[i+1];
    checksum += val;  
  }

  // calc ones comp sum by adding upper 16 bits
  checksum += ((unsigned int) checksum) >> 16;
  // get complement
  return ~checksum & 0xFFFF;
} // end network_calc_ip_checksum

u16 network_calc_udp_checksum (u16 len_udp, unsigned char* src_addr, unsigned char* dest_addr, char padding, unsigned char* buff) {
  u16 prot_udp = 17;
  u16 padd = 0;
  u16 word16;
  u32 sum;
  int i;

  // Find out if the length of data is even or odd number. If odd,
  // add a padding byte = 0 at the end of packet
  if ((padding & 1) == 1) {
    padd = 1;
    buff[len_udp] = 0;
  }

  //initialize sum to zero
  sum = 0;

  // make 16 bit words out of every two adjacent 8 bit words and 
  // calculate the sum of all 16 bit words
  for (i = 0; i < len_udp + padd; i += 2) {
    word16 = ((buff[i] << 8) & 0xFF00) + buff[i + 1];
    sum = sum + (unsigned long) word16;
  }
  
  // add the UDP pseudo header which contains the IP source and destination addresses  
  for (i = 0; i < 4; i += 2) {
    word16 = ((src_addr[i] << 8) & 0xFF00) + src_addr[i + 1];
    sum = sum + word16;
  }
    
  for (i = 0; i < 4; i = i + 2) {
    word16 = ((dest_addr[i] << 8) & 0xFF00) + dest_addr[i + 1];
    sum = sum + word16;
  }
  
  // the protocol number and the length of the UDP packet
  sum = sum + prot_udp + len_udp;

  // keep only the last 16 bits of the 32 bit calculated sum and add the carries
  while (sum >> 16)
    sum = (sum & 0xFFFF) + (sum >> 16);

  // Take the one's complement of sum
  sum = ~sum;

  return ((u16) sum);
} // end network_calc_udp_checksum

// Read a binary file using the altera host-fs and send it over the network    
void network_send_file(const char* filename) {
  char filepath[256];
  snprintf(filepath, sizeof(filepath), "/mnt/host/%s", filename);
  
  FILE* fh = fopen (filepath, "rb");
  if (fh == NULL) {
    printf("Cannot open file: %s\n", filepath);
    exit(1);
  }
  
  fseek(fh, 0, SEEK_END);
  long fileSize = ftell(fh);
  fseek(fh, 0, SEEK_SET); // rewind   
  
  // send a zero then filesize packet
  int zero = 0;
  printf("sending zero packet\n");
  network_send_packet((char*)&zero, sizeof(zero));
  gotosleep(2);    
  printf("sending filesize packet\n");
  network_send_packet((char*)&fileSize, sizeof(fileSize));
        
  // send the file in 1KB increments
  unsigned char buff[1024];
  int i=0;
  for(i=0; i<sizeof(buff); ++i) buff[i] = (char)i;
  size_t bytesRead = 0;
    
  do {
    size_t readSize = fread(buff, 1, sizeof(buff), fh);
          
#ifdef DEBUG          
    for(i = 0; i < sizeof(buff); i++) {
      if(i != 0 && i % 16 == 0) printf("\n");      
      if(i == 0 || i % 16 == 0) printf("%06X: ", (int)(bytesRead+i));
      printf("%02X ", buff[i]);
    }
    printf("\n");
#endif // DEBUG       
    
    if(readSize > 0) network_send_packet(buff, readSize);
    bytesRead += readSize;

#ifndef DEBUG    
    printf("%d bytes sent\n", (int) bytesRead);
#endif
        
  } while(!feof(fh));
  
  fclose (fh);
} // end network_send_file

// big to little endian short
unsigned short network_btoles(unsigned short val) {
  return (val << 8) | (val >> 8);
} // end btoles

// check for ascii values out of range
int containsNonPrintableAscii(const char* msg) {
  int i;
  int bufflen = strlen(msg);
  for(i=0; i<bufflen; ++i) {
    unsigned char c = msg[i];
    if((c < 32 && c != 10) || c > 127) return 1;   
  }
  return 0;
} // containsNonPrintableAscii

void network_interrupt_handler() {
  unsigned int receive_status;
  
  receive_status = ReceivePacket(network_rx_buffer, &network_rx_buffer_length);

  if (receive_status == DMFE_SUCCESS) {

    //printf("\n\nReceive Packet Length = %d\n", network_rx_buffer_length);
    
    if (network_rx_buffer_length >= 14) {
      //  A real Ethernet packet
      if (network_rx_buffer[12] == 8 && network_rx_buffer[13] == 0 &&
          network_rx_buffer_length >= 34) {
        // An IP packet
        if (network_rx_buffer[23] == 0x11) {
          // A UDP packet
          char* msg = network_rx_buffer+HEADER_OFFSET_UDP_DATA;
          if( containsNonPrintableAscii(msg) ) {
             //printf("message contained non printable ascii\n"); 
          } else if (network_rx_buffer_length >= HEADER_OFFSET_UDP_DATA) {        
            int bufflen = strlen(msg)+1;
            char* buff = (char*) malloc(bufflen);
            strncpy(buff, msg, bufflen);
            //printf("Received: %s\n", buff);
          }
        } else {
          //printf("Received non-UDP packet\n");
        }
      } else {
        //printf("Received non-IP packet\n");
      }
    } else {
      //printf("Malformed Ethernet packet\n");
    }

  } else {
    //printf("Error receiving packet\n");
  }

  /* Display the number of interrupts on the LEDs */
  network_interrupt_number++;
  outport(SEG7_DISPLAY_BASE, network_interrupt_number);

  /* Clear the DM9000A ISR: PRS, PTS, ROS, ROOS 4 bits, by RW/C1 */
  dm9000a_iow(ISR, 0x3F);
        
  /* Re-enable DM9000A interrupts */
  dm9000a_iow(IMR, INTR_set);
} // network_interrupt_handler

// Initalize the DM9000 and the Ethernet interrupt handler
void network_init() {
  // copy the constant into a buffer to avoid warning  
  char buffer[6];
  memcpy(buffer, SOURCE_MAC_ADDR, sizeof(buffer)); 
  DM9000_init(buffer);
  network_interrupt_number = 0;
  //alt_irq_register(DM9000A_IRQ, NULL, (void*)network_interrupt_handler); 
} // end network_init

// write the port to the network_udp_packet global
void network_set_source_port(unsigned short port) {
  network_udp_packet[HEADER_OFFSET_SOURCE_PORT_H] = (port & 0xFF00) >> 8;
  network_udp_packet[HEADER_OFFSET_SOURCE_PORT_L] = port & 0xFF;
} // end network_set_source_port

// write the port to the network_udp_packet global
void network_set_dest_port(unsigned short port) {  
  network_udp_packet[HEADER_OFFSET_DEST_PORT_H] = (port & 0xFF00) >> 8;
  network_udp_packet[HEADER_OFFSET_DEST_PORT_L] = port & 0xFF;
} // end network_set_dest_port

unsigned char network_udp_packet[] = {
  // Ethernet MAC header
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00,     //   0-5: Destination MAC address  
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00,     //  6-11: Source MAC address
  0x08, 0x00,                             // 12-13: Packet Type: 0x800 = IP
                          
  // IP Header
  0x45,                //     0-7: version (IPv4), 32 bit words (5 words = 20 bytes)
  0x00,                //    8-15: differentiated services field
                       //   16-18: unused
  0x0F,0x0E,           //   19-31: Total Length: 20 bytes for IP header +
                       //          8 bytes for UDP header + x bytes for payload
                       
  0x00, 0x00,          //   32-47: packet ID
  0x40,                //   48-50: flags (don't fragment)
  //0x00,                //   48-50: flags (none)
  0x00,                //   51-63: fragment offset
  0x40,                //   64-71: time-to-live
  0x11,                //   72-79: protocol: 11 = UDP
  0x00,0x00,           //   80-95: header checksum
  0x00,0x00,0x00,0x00, //  96-127: source IP address
  0x00,0x00,0x00,0x00, // 128-159: destination IP address
                          
  // UDP Header
  0x00,0x00,           //    0-15: source port port (26585: garbage)
  0x00,0x00,           //   16-31: destination port (10000: garbage)
  0x00,0x00,           //   32-47: length (136: 8 for UDP header + 128 for data)
  0x00,0x00           //   48-63: checksum: 0 = none
};
