#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <stdint.h>

#define BUFFER_SIZE 1024
#define MAX_FILE_SIZE 16384
#define PORT 9930
#define SRV_IP "127.0.0.1"

int g_verbose = 0;
char* g_filename = NULL;
char* g_serverAddr = SRV_IP;
unsigned short g_port = PORT;

typedef struct sockaddr SA;

enum { 
	VERBOSE = 1,
	VERY_VERBOSE = 2
};

void diep(char*);
int parseArgs(int, char**);
void displayUsage(const char*);
void sendfile(int, SA*, const char*);
     
int main(int argc, char* argv[]) {
	
	if(!parseArgs(argc, argv)) {
		displayUsage(argv[0]);
		return 1;
	}

	printf("Udpclient sending to %s:%d\n", 	g_serverAddr, (int) g_port);  
  
	struct sockaddr_in si_other;
  int i, slen = sizeof(si_other);
  
	int sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sockfd == -1) diep("socket");
    
  memset((char *) &si_other, 0, sizeof(si_other));
  si_other.sin_family = AF_INET;
  si_other.sin_port = htons(g_port);
  if (inet_aton(g_serverAddr, (struct in_addr*) &si_other.sin_addr)==0) {
    fprintf(stderr, "inet_aton() failed\n");
    exit(1);
  }
  
	sendfile(sockfd, (SA*) &si_other, g_filename);  
   
  int err = close(sockfd);
	if(err) diep("close()");
  return 0;
} // end main

void sendfile(int sockfd, SA* si_other, const char* filename) {
	FILE* fh = fopen(filename, "r");

	if(fh == NULL) {
		fprintf(stderr, "Failed to open file: %s\n", filename);
		exit(1);
	}

	char buffer[BUFFER_SIZE];
	size_t size, bytes_sent = 0;
	int i = 0;

	fseek(fh, 0, SEEK_END);
	int32_t filesize = (int32_t) ftell(fh);
	rewind(fh);

	// send the filesize to the server, convert to standard size for portability
	if(g_verbose == VERY_VERBOSE)
		printf("Sent filesize packet=%d (%d bytes)\n", filesize, 
			(unsigned int) sizeof(filesize));

	size = sendto(sockfd, &filesize, sizeof(int32_t), 0, si_other, sizeof(SA));
	if(size == -1) diep("sendto()");
	bytes_sent += size;

	// transmit the file
	do {
		size = fread(buffer, sizeof(char), sizeof(buffer), fh);

		if(ferror(fh)) {
			diep("sendfile()");
			exit(1);
		}

		if(size == 0) break;

			if (g_verbose == VERY_VERBOSE) {
				int j;
				printf("%07x ", (int)(i*sizeof(buffer)));
				for(j = 0; j < size; j+=2) {
					// print out in same format as od -x
					printf("%02x%02x ", 
							(unsigned char) buffer[j+1], 
							(unsigned char) buffer[j]);
				}
				printf("\n");
			} else if (g_verbose == VERBOSE) {
				printf("Sending packet %d (%d bytes)\n", i+1, (int) size);
			}

		size = sendto(sockfd, buffer, size, 0, si_other, sizeof(SA));
	  if (size == -1) diep("sendto()");
		bytes_sent += size;		
		++i;
	} while(!feof(fh) && bytes_sent < MAX_FILE_SIZE);
  
	printf("Transmitted %d bytes of file '%s'\n", (int) bytes_sent, filename);
} // end sendfile

void diep(char *s) {
  perror(s);
  exit(1);
} // end diep

// parse program arguments, return 0 on failure 1 on success
int parseArgs(int argc, char* argv[]) {
	static struct option long_options[] = {
		{"verbose", no_argument, &g_verbose, 1},
		{"quiet", no_argument, &g_verbose, 0},
		{"port", required_argument, 0, 'p'},
		{0,0,0,0} // terminate the structure
	};

	while(1) {
		int option_index = 0;
		int c = getopt_long(argc, argv, "v::p:", long_options, &option_index);

		if (c == -1) break;

		int port;
		switch (c) {
			case 0:
				/* If this option set a flag, do nothing else now. */
				if (long_options[option_index].flag != 0)	break;
				break;
 
			case 'p':
				port = atoi(optarg);
				if(port <= 0 || port >= 65536) {
					printf("Invalid port value %d\n", port);
					return 0;
				}
				g_port = (unsigned short) port;
				break;

			case 'v':
				if (optarg == NULL)
					g_verbose = VERBOSE;
				else if (strcmp(optarg, "v") == 0) 
					g_verbose = VERY_VERBOSE;
				else
					return 0;
				break;	
	
			case '?':
				/* getopt_long already printed an error message. */
				return 0;
 
			default:
				abort ();
		}
	} // end while
	
	// check that the host and filename are the last remaining items
	if(optind+2 != argc) return 0; 
	g_serverAddr = argv[optind];
	g_filename = argv[optind+1];
	return 1;
} // end parseArgs

void displayUsage(const char* self) {
	fprintf(stderr, "Usage: %s [-v | -vv] [-p port] <host> <filename>\n", self);
} // end displayUsage
