#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <signal.h>
#include <stdint.h>

#define BUFFER_SIZE 2048
#define MAX_FILE_SIZE 16384
#define PORT 9930

int g_abort = 0;
int g_verbose = 0;
char* g_destdir = NULL;
unsigned short g_port = PORT;

enum { 
	VERBOSE = 1,
	VERY_VERBOSE = 2
};

typedef struct sockaddr SA;

void diep(char*);
int parseArgs(int, char**);
void displayUsage(const char*);
void recvfile(int sockfd, SA* si_other);
void sigHandler(int s);

int main(int argc, char* argv[]) {
	struct sigaction sigIntHandler;
	sigIntHandler.sa_handler = sigHandler;
	sigemptyset(&sigIntHandler.sa_mask);
	sigIntHandler.sa_flags = 0;
	sigaction(SIGINT, &sigIntHandler, NULL);
  
	struct sockaddr_in si_me, si_other;
  int sockfd, i;

	if(!parseArgs(argc, argv)) {
		displayUsage(argv[0]);
		return 1;
	}

	printf("Udpserver listening for connections on port %d\n", (int) g_port);  

	if ((sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
    diep("socket");
    
  memset((char *) &si_me, 0, sizeof(si_me));
  si_me.sin_family = AF_INET;
  si_me.sin_port = htons(g_port);
  si_me.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(sockfd, (SA*) &si_me, sizeof(si_me)) == -1) diep("bind");
  
	while(!g_abort) {  
		recvfile(sockfd, (SA*) &si_other);
		printf("\n########################################");
		printf("########################################\n\n");
	}
    
  close(sockfd);
  return 0;
} // end main

// receive a file and write its content to the destination directory
void recvfile(int sockfd, SA* si_other) {
	static int fileNum = 0;
	char filename[1024];
	sprintf(filename, "%s/file%d.jpg", g_destdir, ++fileNum);
	FILE* fh = fopen(filename, "w");

	if(fh == NULL) {
		fprintf(stderr, "Failed to open file: %s\n", filename);
		exit(1);
	}

	char buffer[BUFFER_SIZE];
	size_t size, bytes_recv = 0;
	socklen_t slen = sizeof(SA);
	int i = 0;
	int32_t filesize;

	// first receive an 8 byte message containing the filesize
	size = recvfrom(sockfd, &filesize, sizeof(int32_t), 0, si_other, &slen);
	// convert si_other back to sockaddr_in and get the ip address of client
	struct sockaddr_in* si = (struct sockaddr_in*) si_other;
	
	char* clientAddr = inet_ntoa( (struct in_addr) si->sin_addr);
  if (clientAddr == NULL) {
    fprintf(stderr, "inet_ntoa() failed\n");
    exit(1);
  }

	char filesizeBuff[256];
	
	if(g_verbose == VERBOSE)
		printf("Received size packet from %s 0x%02X 0x%02X 0x%02X 0x%02X\n", 
			clientAddr, 
			(unsigned char) filesizeBuff[0], 
			(unsigned char) filesizeBuff[1],
			(unsigned char) filesizeBuff[2],
			(unsigned char) filesizeBuff[3]
		);
		printf("packet size=%d\n", (unsigned int) size);
		
		//printf("Received size packet from %s (file=%d bytes, packet=%d bytes)\n", 
		//	clientAddr, (unsigned int)filesizeBuff, (unsigned int)size);

	return;
  if (size == -1) diep("recvfrom()");
	if (size != (sizeof(int32_t))) {
		fprintf(stderr, "Protocol error, client did not send filesize (%d)\n",
			(int) size);
		exit(1);
	}

	do {
		size = recvfrom(sockfd, buffer, sizeof(buffer), 0, si_other, &slen);
	  if (size == -1) diep("recvfrom()");
		if (size == 0) break;
		bytes_recv += size;
	
		si = (struct sockaddr_in*) si_other;
	
		clientAddr = inet_ntoa( (struct in_addr) si->sin_addr);
		if (clientAddr == NULL) {
    	fprintf(stderr, "inet_ntoa() failed\n");
	    exit(1);
  	}

		if (g_verbose == VERY_VERBOSE) {
			int j;
			printf("%07x ", (int)(i*sizeof(buffer)));
			for(j = 0; j < size; j+=2) {
				// print out in same format as od -x
				if(j + 1 < size) {
					printf("%02x%02x ", 
							(unsigned char) buffer[j+1], 
							(unsigned char) buffer[j]);
				} else {
					printf("00%02x ", (unsigned char) buffer[j]);
				}
			}
			printf("\n");
		} else if (g_verbose == VERBOSE) {
			printf("Received data packet %d from %s (%d bytes)\n", i+1, 
				clientAddr, (unsigned int) size);
		}

		int res = fwrite(buffer, sizeof(char), size, fh);
		if(res != size) diep("fwrite()");
		
		++i;
	} while(bytes_recv < filesize && bytes_recv < MAX_FILE_SIZE);

	char basefilename[1024];
	sprintf(basefilename, "file%d.jpg", fileNum);
	printf("Received %d bytes, output written to file '%s'\n", 
		(int)bytes_recv, basefilename);  
} // end sendfile

void diep(char *s) {
  perror(s);
  exit(1);
} // end diep

// parse program arguments, return 0 on failure 1 on success
int parseArgs(int argc, char* argv[]) {
	static struct option long_options[] = {
		{"verbose", no_argument, &g_verbose, 1},
		{"quiet", no_argument, &g_verbose, 0},
		{"port", required_argument, 0, 'p'},
		{0,0,0,0} // terminate the structure
	};

	while(1) {
		int option_index = 0;
		int c = getopt_long(argc, argv, "v::p:", long_options, &option_index);

		if (c == -1) break;

		int port;
		switch (c) {
			case 0:
				/* If this option set a flag, do nothing else now. */
				if (long_options[option_index].flag != 0)	break;
				break;
 
			case 'p':
				port = atoi(optarg);
				if(port <= 0 || port >= 65536) {
					printf("Invalid port value %d\n", port);
					return 0;
				}
				g_port = (unsigned short) port;
				break;
	
			case 'v':
				if (optarg == NULL)
					g_verbose = VERBOSE;
				else if (strcmp(optarg, "v") == 0) 
					g_verbose = VERY_VERBOSE;
				else
					return 0;
				break;	
	
			case '?':
				/* getopt_long already printed an error message. */
				return 0;
 
			default:
				abort ();
		}
	} // end while
	
	// check that the filename is the last remaining item
	if(optind+1 != argc) return 0;	

	// strip last '/' from dir
	g_destdir = argv[optind];
	if(g_destdir == NULL) {
		g_destdir = "."; // use the local directory
		return 1;
	}
	int len = strlen(g_destdir);
	if(g_destdir[len-1] == '/') g_destdir[len-1] = '\0';
	return 1;
} // end parseArgs

void displayUsage(const char* self) {
	fprintf(stderr, "Usage: %s [-v] [-p port] [folder]\n", self);
} // end displayUsage

// capture ctrl+c for controlled shutdown
void sigHandler(int s) {
	g_abort = 1;
} // end sigHandler

